"""Test factories."""

from polyfactory.factories.pydantic_factory import ModelFactory

from .models import DeezerSong, DeezerSongResponse


class DeezerSongFactory(ModelFactory[DeezerSong]):
    """DeezerSong factory."""


class DeezerSongResponseFactory(ModelFactory[DeezerSongResponse]):
    """DeezerSongResponse factory."""



"""Dzr configuration."""

import logging
from functools import cache
from pathlib import Path

from pydantic import computed_field
from pydantic.networks import HttpUrl
from pydantic_extra_types.color import Color
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)
from typer import get_app_dir

from .deezer import StreamQuality
from .models import OnzrTheme

logger = logging.getLogger(__name__)

APP_NAME: str = "onzr"
SETTINGS_FILE: Path = Path("settings.yaml")


def get_onzr_dir() -> Path:
    """Get Onzr application directory."""
    return Path(get_app_dir(APP_NAME))


class Settings(BaseSettings):
    """Onzr application settings."""

    DEBUG: bool = False

    # Server
    SCHEMA: str = "http"
    HOST: str = "localhost"
    PORT: int = 9473
    API_ROOT_URL: str = "/api/v1"
    TRACK_STREAM_ENDPOINT: str = "/queue/{rank}/stream"
    PING_TIMEOUT: float = 0.1  # in seconds

    @computed_field  # type: ignore[prop-decorator]
    @property
    def SERVER_BASE_URL(self) -> HttpUrl:
        """Onzr server base URL."""
        return HttpUrl(f"{self.SCHEMA}://{self.HOST}:{self.PORT}{self.API_ROOT_URL}")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def TRACK_STREAM_URL(self) -> str:
        """Onzr server track stream URL."""
        return f"{self.SERVER_BASE_URL}{self.TRACK_STREAM_ENDPOINT}"

    # Customization
    THEME: OnzrTheme = OnzrTheme(
        # Base palette
        primary_color=Color("#9B6BDF"),
        secondary_color=Color("#75D7EC"),
        tertiary_color=Color("#E356A7"),
        # Entities
        title_color=Color("#9B6BDF"),
        artist_color=Color("#75D7EC"),
        album_color=Color("#E356A7"),
        # Messages
        alert_color=Color("red"),
    )

    # Deezer
    QUALITY: StreamQuality = StreamQuality.MP3_128
    DEEZER_BLOWFISH_SECRET: str
    ARL: str

    # Player
    # How long should we wait before getting player status after player control action?
    STATE_DELAY: float = 0.005  # in seconds

    model_config = SettingsConfigDict(
        env_prefix=f"{APP_NAME.upper()}_",
        case_sensitive=True,
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """Add Toml configuration support."""
        return env_settings, YamlConfigSettingsSource(
            settings_cls, yaml_file=get_onzr_dir() / SETTINGS_FILE
        )


@cache
def get_settings() -> Settings:
    """Get settings."""
    logger.debug(f"Loading settings from Onzr directory: {get_onzr_dir()}")
    # ARL and DEEZER_BLOWFISH_SECRET are missing in instantiation since
    # those should be loaded using the YAML configuration
    settings = Settings()  # type: ignore[call-arg]
    logger.debug(f"Settings: {settings=}")
    return settings



"""Tests configuration."""

import importlib
import json
import logging
import tempfile
import threading
from pathlib import Path

import pytest
import requests
import uvicorn
import yaml
from fastapi.testclient import TestClient
from requests.exceptions import ConnectionError
from typer.testing import CliRunner

import onzr
from onzr import cli, config
from onzr.core import Onzr
from onzr.deezer import Track
from tests.factories import DeezerSongFactory, DeezerSongResponseFactory

logger = logging.getLogger(__name__)


@pytest.fixture
def onzr_dir(monkeypatch):
    """Create test application directory."""
    with tempfile.TemporaryDirectory() as app_dir:
        monkeypatch.setattr(config, "get_onzr_dir", lambda: Path(app_dir))
        importlib.reload(cli)
        yield app_dir


@pytest.fixture
def settings_file(onzr_dir):
    """Configured onzr settings file."""
    yield onzr_dir / config.SETTINGS_FILE


@pytest.fixture
def app_configuration(settings_file):
    """Configured onzr app."""
    module_dir = Path(onzr.__file__).parent
    dist = module_dir / settings_file.with_suffix(".yaml.dist").name

    with dist.open() as f:
        test_config = yaml.safe_load(f)
    test_config["ARL"] = "fake-arl"
    test_config["PORT"] = 9474
    with settings_file.open(mode="w") as f:
        yaml.safe_dump(test_config, f, default_style='"')


@pytest.fixture
def settings(app_configuration):
    """Get configured settings."""
    return config.get_settings()


@pytest.fixture
def cli_runner():
    """CLI runner."""
    yield CliRunner()


@pytest.fixture
def configured_cli_runner(app_configuration):
    """Configured CLI runner."""
    yield CliRunner()


@pytest.fixture
def configured_onzr(responses, app_configuration, monkeypatch):
    """Onzr core instance with an initialized Deezer client."""
    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        body=json.dumps(
            {
                "error": {},
                "results": {
                    "checkForm": "secret_token",
                    "USER": {
                        "USER_ID": 666,
                        "MULTI_ACCOUNT": {"ENABLED": False},
                        "BLOG_NAME": "onzr",
                        "OPTIONS": {
                            "license_token": "fake",
                            "web_hq": True,
                            "web_lossless": True,
                            "license_country": "FR",
                        },
                        "SETTING": {"global": {}},
                    },
                },
            }
        ),
        status=200,
        content_type="application/json",
    )
    instance = Onzr()

    yield instance

    instance.player.stop()


@pytest.fixture
def client(configured_onzr):
    """A test client configured for the server."""
    from onzr import server

    def get_configured_onzr():
        return configured_onzr

    # Override dependency to ensure we have got a deezer-mocked Onzr instance
    server.app.dependency_overrides[server.get_onzr] = get_configured_onzr
    yield TestClient(server.app)


@pytest.fixture
def test_server(responses, settings, configured_onzr):
    """A uvicorn test server instance.

    This fixture is costly and should only be run for integration testing such as an
    API client without mocking the server.
    """
    from onzr import server as onzr_server

    def get_configured_onzr():
        return configured_onzr

    # Override dependency to ensure we have got a deezer-mocked Onzr instance
    onzr_server.app.dependency_overrides[onzr_server.get_onzr] = get_configured_onzr

    responses.add_passthru(str(settings.SERVER_BASE_URL))

    config = uvicorn.Config(
        onzr_server.app, host=settings.HOST, port=settings.PORT, log_level="info"
    )
    server = uvicorn.Server(config)

    # Start a real test server in the background
    thread = threading.Thread(
        target=server.run,
        daemon=True,
    )
    thread.start()

    # Wait for the server to start
    max = 150
    attempt = 0
    while (attempt := attempt + 1) <= max:
        try:
            requests.get(settings.SERVER_BASE_URL, timeout=0.1)
        except ConnectionError:
            pass

    yield thread

    # Gracefully shutdown the server
    server.should_exit = True
    thread.join()


@pytest.fixture
def track(responses, configured_onzr, faker, monkeypatch):
    """Track factory fixture."""

    def stream_local_file(_):
        """Stream the same file for every track."""
        chunk_size: int = 2048 * 3
        with Path("./tests/intro-lvs.mp3").open("rb") as content:
            while chunk := content.read(chunk_size):
                yield chunk

    def _track(track_id: int | None = None):
        track_id = faker.pyint() if track_id is None else track_id
        responses.post(
            "http://www.deezer.com/ajax/gw-light.php",
            status=200,
            json=DeezerSongResponseFactory.build(
                error={}, results=DeezerSongFactory.build(SNG_ID=track_id)
            ).model_dump(),
        )
        track = Track(configured_onzr.deezer, track_id)
        monkeypatch.setattr(track, "stream", stream_local_file)
        return track

    return _track



"""Onzr server tests."""

from io import BytesIO
from time import sleep
from unittest.mock import patch

from fastapi import status

from onzr.models import PlayingState

from .factories import DeezerSongFactory, DeezerSongResponseFactory


def test_queue_add_empty(client, responses, configured_onzr):
    """Test the POST /queue/ endpoint."""
    track_ids = []
    assert len(configured_onzr.queue.tracks) == 0

    response = client.post("/queue/", json=track_ids)

    assert response.status_code == status.HTTP_200_OK
    assert response.json()["message"] == f"Added {len(track_ids)} track(s) to queue"
    assert len(configured_onzr.queue.tracks) == 0
    for track, expected in zip(configured_onzr.queue.tracks, track_ids, strict=True):
        assert track.track_id == expected


def test_queue_add(client, responses, configured_onzr):
    """Test the POST /queue/ endpoint."""
    track_ids = [1, 2, 3]
    for track_id in track_ids:
        responses.post(
            "http://www.deezer.com/ajax/gw-light.php",
            status=200,
            json=DeezerSongResponseFactory.build(
                error={}, results=DeezerSongFactory.build(SNG_ID=track_id)
            ).model_dump(),
        )

    assert len(configured_onzr.queue.tracks) == 0

    response = client.post("/queue/", json=track_ids)

    assert response.status_code == status.HTTP_200_OK
    assert response.json()["message"] == f"Added {len(track_ids)} track(s) to queue"
    assert len(configured_onzr.queue.tracks) == len(track_ids)
    for track, expected in zip(configured_onzr.queue.tracks, track_ids, strict=True):
        assert track.track_id == expected


def test_queue_clear(client, configured_onzr, track):
    """Test the DELETE /queue/ endpoint."""
    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Empty queue using the API
    response = client.delete("/queue/")
    state = response.json()
    assert state["player"] == "State.Stopped"
    assert state["queue"] == {"playing": None, "queued": 0}
    assert len(configured_onzr.queue) == 0
    assert configured_onzr.queue.is_empty


def test_queue_clear_empty(client, configured_onzr, track):
    """Test the DELETE /queue/ endpoint when the queue is empty."""
    assert configured_onzr.queue.is_empty

    # Empty queue using the API
    response = client.delete("/queue/")
    state = response.json()
    assert state["player"] == "State.Stopped"
    assert state["queue"] == {"playing": None, "queued": 0}
    assert len(configured_onzr.queue) == 0
    assert configured_onzr.queue.is_empty


def test_queue_list(client, configured_onzr, track):
    """Test the GET /queue/ endpoint."""
    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # List queue using the API
    response = client.get("/queue/")
    queue = response.json()
    assert queue["playing"] is None
    assert len(queue["tracks"]) == len(track_ids)
    for t, id_ in zip(queue["tracks"], track_ids, strict=True):
        assert t["track"]["id"] == id_


def test_queue_list_empty(client, configured_onzr, track):
    """Test the GET /queue/ endpoint when the queue is empty."""
    # List queue using the API
    response = client.get("/queue/")
    queue = response.json()
    assert queue["playing"] is None
    assert len(queue["tracks"]) == 0


def test_stream_track_empty_queue(client):
    """Test the GET /queue/{rank}/stream endpoint when queue is empty."""
    with client.stream("GET", "/queue/10/stream") as response:
        assert response.status_code == status.HTTP_204_NO_CONTENT


def test_stream_track_out_of_range(client, configured_onzr, track):
    """Test the GET /queue/{rank}/stream endpoint when rank is out of range."""
    # Fill queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    with client.stream("GET", "/queue/3/stream") as response:
        assert response.status_code == status.HTTP_404_NOT_FOUND


def test_stream_track(client, responses, configured_onzr, track):
    """Test the GET /queue/{rank}/stream endpoint."""
    # Fill queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Track info refresh request
    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=DeezerSongResponseFactory.build(
            error={}, results=DeezerSongFactory.build(SNG_ID=666)
        ).model_dump(),
    )
    rank = 1
    with client.stream("GET", f"/queue/{rank}/stream") as response:
        assert response.status_code == status.HTTP_200_OK
        song_file = BytesIO()
        for chunk in response.iter_raw():
            song_file.write(chunk)

    # Read the file magic number
    song_file.seek(0)
    magic = song_file.read(3)
    # It should match an mp3 file
    assert magic == b"ID3"

    # Now playing rank should be up to date
    assert configured_onzr.queue.playing == rank


def test_now_playing_empty(client, configured_onzr, track):
    """Test the GET /now endpoint when the queue is empty."""
    response = client.get("/now")
    state = PlayingState(**response.json())
    assert state.player.length == -1
    assert state.player.position == -1.0
    assert state.player.state == "State.NothingSpecial"
    assert state.player.time == -1
    assert state.track is None


def test_now_playing(client, configured_onzr, track):
    """Test the GET /now endpoint."""
    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Play
    configured_onzr.player.play()
    sleep(0.5)

    response = client.get("/now")
    state = PlayingState(**response.json())
    assert state.player.length == 0
    assert state.player.position == 0
    assert state.player.state == "State.Ended"
    assert state.player.time == 0
    assert state.track is None


def test_play(client, configured_onzr, track):
    """Test the POST /play endpoint."""
    with patch.object(configured_onzr.player, "play", return_value=None) as mocked_play:
        response = client.post("/play", json={})
        status = response.json()
        # Queue is empty
        assert status == {
            "action": "play",
            "state": {
                "player": "State.NothingSpecial",
                "queue": {"playing": None, "queued": 0},
            },
        }
        assert mocked_play.called

    # Add tracks to queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])

    # FIXME
    # Start playing
    response = client.post("/play", json={})
    status = response.json()
    assert status == {
        "action": "play",
        "state": {
            "player": "State.Opening",
            "queue": {"playing": None, "queued": 3},
        },
    }


def test_pause(client, configured_onzr):
    """Test the POST /pause endpoint."""
    with patch.object(
        configured_onzr.player, "pause", return_value=None
    ) as mocked_pause:
        response = client.post("/pause")
        status = response.json()
        # Queue is empty
        assert status == {
            "action": "pause",
            "state": {
                "player": "State.NothingSpecial",
                "queue": {"playing": None, "queued": 0},
            },
        }
        assert mocked_pause.called

    # FIXME
    # Should play something first and then pause


def test_stop(client, configured_onzr):
    """Test the POST /stop endpoint."""
    with patch.object(configured_onzr.player, "stop", return_value=None) as mocked_stop:
        response = client.post("/stop")
        status = response.json()
        # Queue is empty
        assert status == {
            "action": "stop",
            "state": {
                "player": "State.NothingSpecial",
                "queue": {"playing": None, "queued": 0},
            },
        }
        assert mocked_stop.called

    # FIXME
    # Should play something first and then stop


def test_next(client, configured_onzr):
    """Test the POST /next endpoint."""
    with patch.object(configured_onzr.player, "next", return_value=None) as mocked_next:
        response = client.post("/next")
        status = response.json()
        # Queue is empty
        assert status == {
            "action": "next",
            "state": {
                "player": "State.NothingSpecial",
                "queue": {"playing": None, "queued": 0},
            },
        }
        assert mocked_next.called

    # FIXME
    # Should play something first and then next


def test_previous(client, configured_onzr, track):
    """Test the POST /previous endpoint."""
    with patch.object(
        configured_onzr.player, "previous", return_value=None
    ) as mocked_previous:
        response = client.post("/previous")
        status = response.json()
        # Queue is empty
        assert status == {
            "action": "previous",
            "state": {
                "player": "State.NothingSpecial",
                "queue": {"playing": None, "queued": 0},
            },
        }
        assert mocked_previous.called

    # FIXME
    # Should play something first and then previous


def test_state(client, configured_onzr, track):
    """Test the GET /state endpoint."""
    response = client.get("/state")
    state = response.json()
    # Queue is empty
    assert state == {
        "player": "State.NothingSpecial",
        "queue": {"playing": None, "queued": 0},
    }

    # Start playing
    configured_onzr.player.play()
    response = client.get("/state")
    state = response.json()
    # Queue is still empty
    assert state == {
        "player": "State.NothingSpecial",
        "queue": {"playing": None, "queued": 0},
    }

    # Add tracks to queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])

    # Start playing for real
    configured_onzr.player.play()
    response = client.get("/state")
    state = response.json()
    assert state == {
        "player": "State.Opening",
        # FIXME
        # Should mock queue.playing
        "queue": {"playing": None, "queued": 3},
    }


def test_ping(client):
    """Test the GET /ping endpoint."""
    response = client.get("/ping")
    assert response.status_code == status.HTTP_200_OK
    assert response.json() is None



"""Onzr: deezer client."""

import functools
import hashlib
import logging
from enum import IntEnum
from threading import Thread
from typing import Any, Generator, Iterator, List, Optional, no_type_check

import deezer
import requests
from Cryptodome.Cipher import Blowfish
from pydantic import HttpUrl

from .exceptions import DeezerTrackException
from .models import (
    AlbumShort,
    ArtistShort,
    Collection,
    StreamQuality,
    TrackInfo,
    TrackShort,
)

logger = logging.getLogger(__name__)


class DeezerClient(deezer.Deezer):
    """A wrapper for the Deezer API client."""

    def __init__(
        self,
        arl: str,
        blowfish: str,
        fast: bool = False,
    ) -> None:
        """Instantiate the Deezer API client.

        Fast login is useful to quicky access some API endpoints such as "search" but
        won't work if you need to stream tracks.
        """
        super().__init__()

        self.arl = arl
        self.blowfish = blowfish
        if fast:
            self._fast_login()
        else:
            self._login()

    def _login(self):
        """Login to deezer API."""
        logger.debug("Login in to deezer using defined ARL…")
        self.login_via_arl(self.arl)

    def _fast_login(self):
        """Fasting login using ARL cookie."""
        cookie_obj = requests.cookies.create_cookie(
            domain=".deezer.com",
            name="arl",
            value=self.arl,
            path="/",
            rest={"HttpOnly": True},
        )
        self.session.cookies.set_cookie(cookie_obj)
        self.logged_in = True

    @staticmethod
    def _to_tracks(data) -> Generator[TrackShort, None, None]:
        """API results to TrackShort."""
        for track in data:
            yield TrackShort(
                id=track.get("id"),
                title=track.get("title"),
                album=track.get("album").get("title"),
                artist=track.get("artist").get("name"),
            )

    @staticmethod
    def _to_albums(data, artist: ArtistShort) -> Generator[AlbumShort, None, None]:
        """API results to AlbumShort."""
        for album in data:
            logger.debug(f"{album=}")
            yield AlbumShort(
                id=album.get("id"),
                title=album.get("title"),
                release_date=album.get("release_date"),
                artist=artist.name,
            )

    def artist(
        self,
        artist_id: int,
        radio: bool = False,
        top: bool = True,
        albums: bool = False,
        limit: int = 10,
    ) -> List[TrackShort] | List[AlbumShort]:
        """Get artist tracks."""
        response = self.api.get_artist(artist_id)
        artist = ArtistShort(id=response.get("id"), name=response.get("name"))
        logger.debug(f"{artist=}")

        if radio:
            response = self.api.get_artist_radio(artist_id, limit=limit)
            return list(self._to_tracks(response["data"]))
        elif top:
            response = self.api.get_artist_top(artist_id, limit=limit)
            return list(self._to_tracks(response["data"]))
        elif albums:
            response = self.api.get_artist_albums(artist_id, limit=limit)
            return list(self._to_albums(response["data"], artist))
        else:
            raise ValueError(
                "Either radio, top or albums should be True to get artist details"
            )

    def album(self, album_id: int) -> List[TrackShort]:
        """Get album tracks."""
        response = self.api.get_album(album_id)
        logger.debug(f"{response=}")
        return list(self._to_tracks(response["tracks"]["data"]))

    def search(
        self,
        artist: str = "",
        album: str = "",
        track: str = "",
        strict: bool = False,
    ) -> Collection:
        """Mixed custom search."""
        results: Collection = []

        if len(list(filter(None, (artist, album, track)))) > 1:
            response = self.api.advanced_search(
                artist=artist, album=album, track=track, strict=strict
            )
            results = list(self._to_tracks(response["data"]))
        elif artist:
            response = self.api.search_artist(artist)
            results = [
                ArtistShort(
                    id=a.get("id"),
                    name=a.get("name"),
                )
                for a in response["data"]
            ]
        elif album:
            response = self.api.search_album(album)
            results = [
                AlbumShort(
                    id=a.get("id"),
                    title=a.get("title"),
                    release_date=a.get("release_date"),
                    artist=a.get("artist").get("name"),
                )
                for a in response["data"]
            ]
        elif track:
            response = self.api.search_track(track)
            results = list(self._to_tracks(response["data"]))

        return results


class TrackStatus(IntEnum):
    """Track statuses."""

    IDLE = 1
    STREAMING = 2
    STREAMED = 3


class AlbumCoverSize(IntEnum):
    """Album cover sizes."""

    SMALL = 0
    MEDIUM = 1
    BIG = 2
    XL = 3


def get_album_cover_filename(size: AlbumCoverSize) -> str:
    """Get album cover filename given its size."""
    match size:
        case AlbumCoverSize.SMALL:
            return "56x56-000000-80-0-0.jpg"
        case AlbumCoverSize.MEDIUM:
            return "250x250-000000-80-0-0.jpg"
        case AlbumCoverSize.BIG:
            return "500x500-000000-80-0-0.jpg"
        case AlbumCoverSize.XL:
            return "1000x1000-000000-80-0-0.jpg"


class Track:
    """A Deezer track."""

    def __init__(
        self,
        client: DeezerClient,
        track_id: int,
        background: bool = False,
    ) -> None:
        """Instantiate a new track."""
        self.deezer = client
        self.track_id = track_id
        self.session = requests.Session()

        self.track_info: Optional[TrackInfo] = None
        # Fetch track info in a separated thread to make instantiation non-blocking
        if background:
            thread = Thread(target=self._set_track_info)
            thread.start()
        else:
            self._set_track_info()

        self.key: bytes = self._generate_blowfish_key()
        self.status: TrackStatus = TrackStatus.IDLE
        self.streamed: int = 0

    def _set_track_info(self):
        """Get track info."""
        track_info = self.deezer.gw.get_track(self.track_id)
        logger.debug("Track info: %s", track_info)
        filesizes = {
            "FILESIZE_MP3_128": StreamQuality.MP3_128,
            "FILESIZE_MP3_320": StreamQuality.MP3_320,
            "FILESIZE_FLAC": StreamQuality.FLAC,
        }
        self.track_info = TrackInfo(
            id=track_info["SNG_ID"],
            token=track_info["TRACK_TOKEN"],
            duration=track_info["DURATION"],
            artist=track_info["ART_NAME"],
            title=(
                f"{track_info['SNG_TITLE']} {track_info['VERSION']}"
                if "VERSION" in track_info and track_info["VERSION"]
                else track_info["SNG_TITLE"]
            ),
            album=track_info["ALB_TITLE"],
            picture=track_info["ALB_PICTURE"],
            formats=[
                filesizes[size] for size in filesizes if int(track_info[size]) > 0
            ],
        )
        if not len(self.formats):
            raise DeezerTrackException(
                f"No available formats detected for track {self.track_id}"
            )
        logger.debug(f"{self.track_info}")

    def refresh(self):
        """Refresh track info."""
        logger.debug("Refreshing track info…")
        self._set_track_info()

    def _get_url(self, quality: StreamQuality) -> str:
        """Get URL of the track to stream."""
        logger.debug(f"Getting track url with quality {quality}…")
        url = self.deezer.get_track_url(self.token, quality.value)
        return url

    def _generate_blowfish_key(self) -> bytes:
        """Generate the blowfish key for Deezer downloads.

        Taken from: https://github.com/nathom/streamrip/
        """
        md5_hash = hashlib.md5(str(self.track_id).encode()).hexdigest()  # noqa: S324
        # good luck :)
        return "".join(
            chr(functools.reduce(lambda x, y: x ^ y, map(ord, t)))
            for t in zip(
                md5_hash[:16],
                md5_hash[16:],
                self.deezer.blowfish,
                strict=False,
            )
        ).encode()

    def _decrypt(self, chunk):
        """Decrypt blowfish encrypted chunk."""
        return Blowfish.new(  # noqa: S304
            self.key,
            Blowfish.MODE_CBC,
            b"\x00\x01\x02\x03\x04\x05\x06\x07",
        ).decrypt(chunk)

    def _get_track_info_attribute(self, field: str) -> Any | None:
        """Get self.track_info attribute if defined."""
        if self.track_info is None:
            return "fetching…"
        return getattr(self.track_info, field, None)

    @property
    def token(self) -> str | None:
        """Get track token."""
        return self._get_track_info_attribute("token")

    @property
    def duration(self) -> int | None:
        """Get track duration (in seconds)."""
        return self._get_track_info_attribute("duration")

    @property
    def artist(self) -> str | None:
        """Get track artist."""
        return self._get_track_info_attribute("artist")

    @property
    def title(self) -> str | None:
        """Get track title."""
        return self._get_track_info_attribute("title")

    @property
    def album(self) -> str | None:
        """Get track album."""
        return self._get_track_info_attribute("album")

    @property
    def formats(self) -> List[StreamQuality]:
        """Get track formats."""
        return self._get_track_info_attribute("formats") or []

    def query_quality(self, quality: StreamQuality) -> StreamQuality:
        """Get track quality among available formats.

        All stream qualities are not available for every tracks, if queried quality
        is not available return the best available quality among supported formats.
        """
        if quality in self.formats:
            return quality
        return self.formats[-1]

    @property
    def picture(self) -> str | None:
        """Get track picture."""
        return self._get_track_info_attribute("picture")

    def _cover(self, size: AlbumCoverSize) -> HttpUrl | None:
        """Get track album cover URL given requested size."""
        return (
            HttpUrl(
                "https://e-cdns-images.dzcdn.net/images/cover/"
                f"{self.picture}/"
                f"{get_album_cover_filename(size)}"
            )
            if self.picture
            else None
        )

    @property
    def cover_small(self) -> HttpUrl | None:
        """Get small album cover URL."""
        return self._cover(AlbumCoverSize.SMALL)

    @property
    def cover_medium(self) -> HttpUrl | None:
        """Get medium album cover URL."""
        return self._cover(AlbumCoverSize.MEDIUM)

    @property
    def cover_big(self) -> HttpUrl | None:
        """Get big album cover URL."""
        return self._cover(AlbumCoverSize.BIG)

    @property
    def cover_xl(self) -> HttpUrl | None:
        """Get XL album cover URL."""
        return self._cover(AlbumCoverSize.XL)

    @property
    def full_title(self) -> str:
        """Get track full title (artist/title/album)."""
        return f"{self.artist} - {self.title} [{self.album}]"

    def stream(self, quality: StreamQuality = StreamQuality.MP3_128) -> Iterator[bytes]:
        """Fetch track in-memory.

        quality (StreamQuality): audio file to stream quality
        """
        if (best := self.query_quality(quality)) != quality:
            logger.warning(
                (
                    "Required track quality %s is not available. Will try best "
                    "available format instead: %s"
                ),
                quality,
                best,
            )
            quality = best

        logger.debug(
            "Start streaming track: "
            f"▶️ {self.full_title} (ID: {self.track_id} Q: {quality})"
        )

        chunk_sep = 2048
        chunk_size = 3 * chunk_sep
        self.streamed = 0
        self.status = TrackStatus.IDLE

        url = self._get_url(quality)
        with self.session.get(url, stream=True) as r:
            r.raise_for_status()
            filesize = int(r.headers.get("Content-Length", 0))
            logger.debug(f"Track size: {filesize}")
            self.status = TrackStatus.STREAMING

            for chunk in r.iter_content(chunk_size):
                if len(chunk) > chunk_sep:
                    dchunk = self._decrypt(chunk[:chunk_sep]) + chunk[chunk_sep:]
                else:
                    dchunk = chunk
                self.streamed += chunk_size
                yield dchunk

        # We are done here
        self.status = TrackStatus.STREAMED
        logger.debug(f"Track fully streamed {self.streamed}")

    # Pydantic will raise an error for us
    @no_type_check
    def serialize(self) -> TrackShort:
        """Serialize current track."""
        return TrackShort(
            id=self.track_id,
            title=self.title,
            album=self.album,
            artist=self.artist,
        )



"""Onzr: command line interface."""

import json
import logging
import sys
import time
from datetime import date
from enum import IntEnum
from functools import cache, wraps
from importlib.metadata import version as import_lib_version
from pathlib import Path
from random import shuffle
from typing import List, cast

import click
import pendulum
import typer
import uvicorn
import yaml
from rich.console import Console, Group
from rich.live import Live
from rich.logging import RichHandler
from rich.progress_bar import ProgressBar
from rich.prompt import Prompt
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text
from typing_extensions import Annotated
from uvicorn.config import LOG_LEVELS

from onzr.exceptions import OnzrConfigurationError

from .client import OnzrClient
from .config import (
    SETTINGS_FILE,
    get_onzr_dir,
    get_settings,
)
from .deezer import DeezerClient
from .models import (
    AlbumShort,
    ArtistShort,
    Collection,
    PlayerControl,
    ServerState,
    TrackShort,
)

FORMAT = "%(message)s"
logging_console = Console(stderr=True)
logging_config = {
    "level": logging.INFO,
    "format": FORMAT,
    "datefmt": "[%X]",
    "handlers": [RichHandler(console=logging_console)],
}
logging.basicConfig(**logging_config)  # type: ignore[arg-type]

cli = typer.Typer(name="onzr", no_args_is_help=True, pretty_exceptions_short=True)
console = Console()
logger = logging.getLogger(__name__)


@cache
def get_theme():
    """Get Onzr theme."""
    return get_settings().THEME


class ExitCodes(IntEnum):
    """Onzr exit codes."""

    OK = 0
    INCOMPLETE_CONFIGURATION = 10
    INVALID_CONFIGURATION = 11
    INVALID_ARGUMENTS = 20
    NOT_FOUND = 30
    SERVER_DOWN = 40


def get_deezer_client(quiet: bool = False) -> DeezerClient:
    """Get Deezer client for simple API queries."""
    settings = get_settings()

    if not quiet:
        console.print("🚀 login in to Deezer…", style="cyan")

    return DeezerClient(
        arl=settings.ARL,
        blowfish=settings.DEEZER_BLOWFISH_SECRET,
        fast=True,
    )


def require_server(func):
    """A command decorator that tests if Onzr server is running."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        client = OnzrClient()
        theme = get_theme()
        if not client.ping():
            console.print(
                f"[{theme.alert_color}]❌ "
                "Onzr server is down, run `onzr serve` first."
                f"[/{theme.alert_color}]"
            )
            raise typer.Exit(ExitCodes.SERVER_DOWN)
        func(*args, **kwargs)

    return wrapper


def print_collection_ids(collection: Collection):
    """Print a collection as ids."""
    for item in collection:
        console.print(item.id)


def print_collection_table(collection: Collection, title="Collection"):
    """Print a collection as a table."""
    theme = get_theme()
    table = Table(title=title)

    sample = collection[0]
    show_artist = (
        True
        if isinstance(sample, TrackShort)
        or isinstance(sample, AlbumShort)
        or isinstance(sample, ArtistShort)
        else False
    )
    show_album = (
        True
        if isinstance(sample, TrackShort) or isinstance(sample, AlbumShort)
        else False
    )
    show_track = True if isinstance(sample, TrackShort) else False
    show_release = True if isinstance(sample, AlbumShort) else False
    logger.debug(f"{show_artist=} - {show_album=} - {show_track=}")

    table.add_column("ID", justify="right")
    if show_track:
        table.add_column("Track", style=theme.title_color.as_hex())
    if show_album:
        table.add_column("Album", style=theme.album_color.as_hex())
    if show_artist:
        table.add_column("Artist", style=theme.artist_color.as_hex())
    if show_release:
        table.add_column("Released")

    # Sort albums by release date
    # FIXME: mypy does not get that we are dealing with a List[AlbumShort] collection
    if isinstance(sample, AlbumShort):
        albums_with_release_date = set(
            filter(lambda x: x.release_date is not None, collection)  # type: ignore[attr-defined]
        )
        albums_without_release_date = list(set(collection) - albums_with_release_date)
        collection = sorted(
            albums_with_release_date,
            key=lambda i: date.fromisoformat(i.release_date),  # type: ignore[attr-defined]
            reverse=True,
        )  # type: ignore[assignment]
        collection.extend(albums_without_release_date)  # type: ignore[arg-type]

    for item in collection:
        table.add_row(*map(str, item.model_dump().values()))

    console.print(table)


@cli.command()
def init():
    """Intialize onzr player."""
    console.print("⚙️ Initializing onzr…")

    app_dir = get_onzr_dir()
    module_dir = Path(__file__).parent

    # Create Onzr config directory if needed
    logger.debug(f"Creating application directory: {app_dir}")
    app_dir.mkdir(mode=0o755, parents=True, exist_ok=True)

    # Copy original dist
    logger.debug("Will copy distributed configurations…")
    src = module_dir / SETTINGS_FILE.with_suffix(".yaml.dist")
    dest = app_dir / SETTINGS_FILE
    logger.debug(f"{src=} -> {dest=}")

    if dest.exists():
        raise OnzrConfigurationError(f"Configuration file '{dest}' already exists!")

    logger.info(f"Will create base setting file '{dest}'")
    dest.write_text(src.read_text())
    logger.debug(f"Copied base setting file to: {dest}")

    # Open base configuration
    with src.open() as f:
        user_settings = yaml.safe_load(f)

    logger.debug("ARL value will be (re)set.")
    user_settings["ARL"] = Prompt.ask("Paste your ARL 📋")

    logger.info(f"Writing settings configuration to: {dest}")
    with dest.open(mode="w") as f:
        yaml.dump(user_settings, f)

    console.print("🎉 Everything looks ok from here. You can start playing 💫")


@cli.command()
def config(
    path: Annotated[
        bool, typer.Option("--path", "-p", help="Show configuration path and exit.")
    ] = False,
    edit: Annotated[
        bool, typer.Option("--edit", "-e", help="Edit configuration in $EDITOR.")
    ] = False,
):
    """Display or edit Onzr's configuration."""
    user_config_path = get_onzr_dir() / SETTINGS_FILE

    if not user_config_path.exists():
        console.print(
            "[red]Configuration file does not exist, use `onzr init` first.[/red]"
        )
        raise typer.Exit(ExitCodes.INCOMPLETE_CONFIGURATION)

    if path:
        console.print(user_config_path)
        raise typer.Exit(0)

    if edit:
        click.edit(filename=str(user_config_path))
        raise typer.Exit(0)

    with user_config_path.open() as f:
        user_config = f.read()
    console.print(Syntax(user_config, "yaml"))


@cli.command()
def search(  # noqa: PLR0913
    artist: Annotated[
        str, typer.Option("--artist", "-A", help="Search by artist name.")
    ] = "",
    album: Annotated[
        str, typer.Option("--album", "-a", help="Search by album name.")
    ] = "",
    track: Annotated[
        str, typer.Option("--track", "-t", help="Search by track title.")
    ] = "",
    strict: Annotated[
        bool, typer.Option("--strict", "-s", help="Only consider strict matches.")
    ] = False,
    first: Annotated[
        bool, typer.Option("--first", "-f", help="Limit to the first match.")
    ] = False,
    quiet: Annotated[bool, typer.Option("--quiet", "-q", help="Quiet output.")] = False,
    ids: Annotated[
        bool, typer.Option("--ids", "-i", help="Show only result IDs.")
    ] = False,
):
    """Search tracks, artists and/or albums.

    Note that search criterion can be combined (e.g. artist and album).
    """
    if ids:
        quiet = True
    deezer = get_deezer_client(quiet=quiet)
    theme = get_theme()

    if not quiet:
        console.print("🔍 start searching…")
    results = deezer.search(artist, album, track, strict)

    if not results:
        console.print(f"❌ [{theme.alert_color}]No match found[/{theme.alert_color}]")
        raise typer.Exit(code=ExitCodes.NOT_FOUND)

    if first:
        results = results[:1]

    if ids:
        print_collection_ids(results)
        return

    print_collection_table(results, title="Search results")


@cli.command()
def artist(  # noqa: PLR0913
    artist_id: str,
    top: Annotated[
        bool, typer.Option("--top/--no-top", "-t/-T", help="Show artist top tracks.")
    ] = True,
    radio: Annotated[
        bool,
        typer.Option("--radio", "-r", help="Show artist-inspired tracks."),
    ] = False,
    albums: Annotated[
        bool, typer.Option("--albums", "-a", help="Show artist albums.")
    ] = False,
    limit: Annotated[
        int, typer.Option("--limit", "-l", help="Limit to the l first hits.")
    ] = 10,
    quiet: Annotated[bool, typer.Option("--quiet", "-q", help="Quiet output.")] = False,
    ids: Annotated[
        bool, typer.Option("--ids", "-i", help="Show only result IDs.")
    ] = False,
):
    """Get artist popular track ids.

    Remember to increase the default limit to show all artist albums if it has produced
    more than one.
    """
    if all([not top, not radio, not albums]):
        console.print("You should choose either top titles, artist radio or albums.")
        raise typer.Exit(code=ExitCodes.INVALID_ARGUMENTS)
    elif albums:
        top = False
        radio = False

    if ids:
        quiet = True

    if artist_id == "-":
        logger.debug("Reading artist id from stdin…")
        artist_id = click.get_text_stream("stdin").read().strip()
        logger.debug(f"{artist_id=}")

    deezer = get_deezer_client(quiet=quiet)
    collection = deezer.artist(
        int(artist_id), radio=radio, top=top, albums=albums, limit=limit
    )

    if ids:
        print_collection_ids(collection)
        return

    print_collection_table(collection, title="Artist collection")


@cli.command()
def album(
    album_id: str,
    quiet: Annotated[bool, typer.Option("--quiet", "-q", help="Quiet output.")] = False,
    ids: Annotated[
        bool, typer.Option("--ids", "-i", help="Show only result IDs.")
    ] = False,
):
    """Get album tracks."""
    if ids:
        quiet = True

    if album_id == "-":
        logger.debug("Reading artist id from stdin…")
        album_id = click.get_text_stream("stdin").read().strip()
        logger.debug(f"{album_id=}")

    deezer = get_deezer_client(quiet=quiet)
    collection = deezer.album(int(album_id))

    if ids:
        print_collection_ids(collection)
        return

    print_collection_table(collection, title="Album tracks")


@cli.command()
def mix(
    artist: list[str],
    deep: Annotated[
        bool,
        typer.Option(
            "--deep", "-d", help="Create a mix with related artists (like a radio)."
        ),
    ] = False,
    limit: Annotated[
        int, typer.Option("--limit", "-l", help="Limit to the l first hits per artist.")
    ] = 10,
    quiet: Annotated[bool, typer.Option("--quiet", "-q", help="Quiet output.")] = False,
    ids: Annotated[
        bool, typer.Option("--ids", "-i", help="Show only result IDs.")
    ] = False,
):
    """Create a playlist from multiple artists."""
    if ids:
        quiet = True

    deezer = get_deezer_client(quiet=quiet)
    tracks: List[TrackShort] = []

    if not quiet:
        console.print("🍪 cooking the mix…")

    for artist_ in artist:
        result = deezer.search(artist_, strict=True)
        # We expect the search engine to be relevant 🤞
        artist_id = result[0].id
        tracks += cast(
            List[TrackShort],
            deezer.artist(artist_id, radio=deep, top=True, limit=limit),
        )
    shuffle(tracks)

    if ids:
        print_collection_ids(tracks)
        return

    print_collection_table(tracks, title="Onzr Mix tracks")


@cli.command()
@require_server
def add(track_ids: List[str]):
    """Add one (or more) tracks to the queue."""
    if track_ids == ["-"]:
        logger.debug("Reading track ids from stdin…")
        track_ids = click.get_text_stream("stdin").read().split()
        logger.debug(f"{track_ids=}")

    console.print("➕ Adding tracks to queue…")

    client = OnzrClient()
    response = client.queue_add(track_ids)

    console.print(f"✅ {response.message}")


def _client_request(name: str, **kwargs):
    """A generic wrapper that executes a client method."""
    client = OnzrClient()
    method = getattr(client, name)
    response = method(**kwargs)
    return response


@cli.command()
@require_server
def queue():
    """List queue tracks."""
    theme = get_theme()
    queue = _client_request("queue_list")
    if not len(queue):
        console.print(
            "⚠ [yellow]Queue is empty, use [magenta]onzr add[/magenta] "
            "to start adding tracks.[/yellow]"
        )
        raise typer.Exit(0)

    with console.pager(styles=True):
        for qt in queue.tracks:
            track_infos = (
                f"[white][bold]{qt.position + 1:-3d}[/] "
                f"[{theme.title_color}]{qt.track.title}[white] - "
                f"[{theme.artist_color}]{qt.track.artist} "
                f"[{theme.album_color}]({qt.track.album})"
            )
            if queue.playing is not None and qt.position < queue.playing:
                s = f"🏁 [italic]{track_infos}[/italic]"
            elif qt.current:
                s = f"▶  [bold]{track_infos}[/bold]"
            else:
                s = f"🧵 {track_infos}"
            s += "[white]"
            console.print(s)


def _print_server_state(state: ServerState):
    """Print server state."""
    theme = get_theme()
    playing = state.queue.playing + 1 if state.queue.playing is not None else None
    s = (
        "📢 "
        f"Player: [{theme.secondary_color}]{state.player.split('.')[1]}[white]"
        " · "
        f"Queue: [{theme.tertiary_color}]{playing}[white]"
        " / "
        f"[{theme.primary_color}]{state.queue.queued}[white]"
    )
    console.print(s)


def _print_player_control(control: PlayerControl):
    """Print player control action."""
    theme = get_theme()
    match control.action:
        case "play":
            icon = "▶️"
        case "pause":
            icon = "⏯️"
        case "stop":
            icon = "⏹️"
        case "next":
            icon = "⏭"
        case "previous":
            icon = "⏮"
        case _:
            icon = "⁉️"
    console.print(f"{icon}  Action: [{theme.secondary_color}]{control.action}[white] ")
    _print_server_state(control.state)


@cli.command()
@require_server
def clear():
    """Empty queue."""
    state = _client_request("queue_clear")
    _print_server_state(state)


@cli.command()
@require_server
def now(
    follow: Annotated[
        bool, typer.Option("--follow", "-f", help="Follow what's happening.")
    ] = False,
):
    """Show details about the track that is being played and the player status.

    In follow mode, you won't get the prompt back. You should type CTRL+C to exit this
    mode.
    """
    client = OnzrClient()
    theme = get_theme()

    def timecode(duration: pendulum.Duration) -> str:
        """Convert a duration (in ms) to a time code."""
        return (
            f"{duration.hours:02d}:"
            f"{duration.minutes:02d}:"
            f"{duration.remaining_seconds:02d}"
        )

    def get_track_infos(track: TrackShort) -> str:
        """Generate a fully qualified track string."""
        return (
            f"[{theme.title_color}]{track.title} - "
            f"[{theme.artist_color}]{track.artist} - "
            f"[{theme.album_color}]{track.album}"
        )

    def display() -> Group:
        """Now playing."""
        now_playing = client.now_playing()
        queue = client.queue_list()
        track = now_playing.track
        player = now_playing.player
        next_track = (
            queue.tracks[queue.playing + 1].track
            if queue.playing is not None and queue.playing < len(queue) - 1
            else None
        )

        match player.state:
            case "State.Playing":
                icon = "▶️"
            case "State.Paused":
                icon = "⏯️"
            case "State.Stopped":
                icon = "⏹️"
            case "State.Ended":
                icon = "🏁"
            case "State.NothingSpecial":
                icon = "🤷"
            case "State.Opening":
                icon = "📂"
            case "State.Buffering":
                icon = "🌐"
            case _:
                icon = "⁉️"

        track_infos = f"{icon} "
        if track is not None:
            track_infos += get_track_infos(track)
        track_infos += (
            "[white] · "
            f"({queue.playing + 1 if queue.playing is not None else '-'}/{len(queue)})"
        )
        track_duration = pendulum.duration(seconds=player.length / 1000.0)
        track_played = pendulum.duration(seconds=player.time / 1000.0)
        track_played_timecode = Text(
            f"{timecode(track_played)} ", style=theme.secondary_color.as_hex(), end=""
        )
        track_total_timecode = f" [{theme.primary_color}]{timecode(track_duration)}"
        progress_bar = ProgressBar(
            total=player.length,
            completed=player.time,
            complete_style=theme.tertiary_color.as_hex(),
            finished_style=theme.secondary_color.as_hex(),
            width=62,
        )
        coming_next = "Next: "
        if next_track is not None:
            coming_next += get_track_infos(next_track)
        else:
            coming_next += "❎ [italic]Nothing more has been queued[/italic]"

        return Group(
            track_infos,
            track_played_timecode,
            progress_bar,
            track_total_timecode,
            coming_next,
        )

    if not follow:
        console.print(display())
        return

    with Live(display(), refresh_per_second=4) as live:
        while True:
            time.sleep(0.1)
            live.update(display())


@cli.command()
@require_server
def play(
    rank: Annotated[
        int | None,
        typer.Option(
            "--rank", "-r", help="Start playing the queue starting at the rank r."
        ),
    ] = None,
):
    """Play queued tracks.

    If the player is paused, this command will resume the track.
    """
    theme = get_theme()
    if rank is not None and rank < 1:
        console.print(
            (
                "🙈 "
                f"[{theme.alert_color} bold]Invalid rank![/{theme.alert_color} bold] "
                "It should be greater than 0."
            )
        )
        raise typer.Exit(ExitCodes.INVALID_ARGUMENTS)
    control = _client_request("play", rank=rank - 1 if rank else None)
    _print_player_control(control)


@cli.command()
@require_server
def pause():
    """Pause/resume playing."""
    control = _client_request("pause")
    _print_player_control(control)


@cli.command()
@require_server
def stop():
    """Stop playing queue."""
    control = _client_request("stop")
    _print_player_control(control)


@cli.command()
@require_server
def next():
    """Play next track in queue."""
    control = _client_request("next")
    _print_player_control(control)


@cli.command()
@require_server
def previous():
    """Play previous track in queue."""
    control = _client_request("previous")
    _print_player_control(control)


@cli.command()
def serve(
    host: Annotated[
        str, typer.Option("--host", "-H", help="Server host name.")
    ] = "localhost",
    port: Annotated[int, typer.Option("--port", "-P", help="Server port.")] = 9473,
    log_level: Annotated[
        str,
        typer.Option(
            "--log-level",
            "-L",
            help="Server log level (debug, info, warning, error, critical).",
        ),
    ] = "info",
):
    """Run onzr http server."""
    theme = get_theme()
    # Typer does not support complex types such as Litteral, so let's check log_level
    # validity by ourselves.
    allowed_levels: list[str] = ["debug", "info", "warning", "error", "critical"]
    if log_level not in allowed_levels:
        console.print(
            (
                "🙈 "
                f"[{theme.alert_color} bold]"
                "Forbidden log-level!"
                f"[/{theme.alert_color} bold] "
                f"Should be in: {allowed_levels}"
            )
        )
        raise typer.Exit(ExitCodes.INVALID_ARGUMENTS)

    level = LOG_LEVELS[log_level]
    logging_config.update({"level": level})
    logging.basicConfig(**logging_config, force=True)  # type: ignore[arg-type]

    settings = get_settings()
    config = uvicorn.Config(
        "onzr.server:app",
        host=host or settings.HOST,
        port=port or settings.PORT,
        log_level=level,
    )
    server = uvicorn.Server(config)
    server.run()


@cli.command()
@require_server
def state():
    """Get server state."""
    client = OnzrClient()
    state = client.state()
    _print_server_state(state)


@cli.command()
def version():
    """Get program version."""
    theme = get_theme()
    console.print(
        f"🔖 Version: [{theme.secondary_color}]{import_lib_version('onzr')}[white]"
    )


@cli.command()
@require_server
def openapi():
    """Get Onzr HTTP API OpenAPI schema."""
    from onzr.server import app

    sys.stdout.write(f"{json.dumps(app.openapi())}\n")



"""Onzr: Pydantic models."""

from enum import StrEnum
from typing import Annotated, List, Optional, TypeAlias

from pydantic import BaseModel, Field
from pydantic_extra_types.color import Color


class OnzrTheme(BaseModel):
    """Onzr theme."""

    primary_color: Color
    secondary_color: Color
    tertiary_color: Color
    title_color: Color
    artist_color: Color
    album_color: Color
    alert_color: Color


class QueueState(BaseModel):
    """Queue state."""

    playing: int | None
    queued: int


class ServerState(BaseModel):
    """Onzr server state."""

    # Does not support VLC Enums
    player: str
    queue: QueueState


class StreamQuality(StrEnum):
    """Track stream quality."""

    MP3_128 = "MP3_128"
    MP3_320 = "MP3_320"
    FLAC = "FLAC"

    @property
    def media_type(self) -> str:
        """Get media type corresponding to selected quality."""
        if self == StreamQuality.FLAC:
            return "audio/flac"
        return "audio/mpeg"


class PlayerControl(BaseModel):
    """Player controls."""

    action: str
    state: ServerState


class ServerMessage(BaseModel):
    """Generic server message."""

    message: str


class ArtistShort(BaseModel):
    """A small model to represent an artist."""

    id: int
    name: str


class AlbumShort(BaseModel):
    """A small model to represent an artist."""

    id: int
    title: str
    artist: Optional[str] = None
    release_date: Optional[str] = None

    def __hash__(self):
        """Make AlbumShort hashable."""
        return hash(self.id)


class TrackShort(BaseModel):
    """A small model to represent an artist."""

    id: int
    title: str
    album: str
    artist: str


class TrackInfo(BaseModel):
    """Used track data."""

    id: int
    title: str
    album: str
    artist: str
    picture: str
    token: str
    duration: int
    formats: List[StreamQuality]


Collection: TypeAlias = List[ArtistShort] | List[AlbumShort] | List[TrackShort]


class QueuedTrack(BaseModel):
    """Queued track."""

    current: bool
    position: int
    track: TrackShort


class QueuedTracks(BaseModel):
    """Queued Tracks list."""

    playing: int | None
    tracks: List[QueuedTrack]

    def __len__(self):
        """Get tracks length."""
        return len(self.tracks)


class PlayerState(BaseModel):
    """Detailled player state."""

    state: str
    length: int = 0
    time: int = 0
    position: float = 0.0


class PlayingState(BaseModel):
    """Playing player state."""

    player: PlayerState
    track: Optional[TrackShort] = None


class PlayQueryParams(BaseModel):
    """Play endpoint parameters."""

    rank: Optional[Annotated[int, Field(strict=True, ge=0)]] = None



"""Onzr deezer tests."""

import pytest
from pydantic import HttpUrl

from onzr.deezer import StreamQuality, Track, TrackStatus
from onzr.exceptions import DeezerTrackException
from onzr.models import TrackInfo, TrackShort
from tests.factories import DeezerSongFactory, DeezerSongResponseFactory


def test_stream_quality_enum():
    """Test the StreamQuality enum."""
    assert StreamQuality.FLAC.media_type == "audio/flac"
    assert StreamQuality.MP3_320.media_type == "audio/mpeg"
    assert StreamQuality.MP3_128.media_type == "audio/mpeg"


def test_track_init(configured_onzr, responses):
    """Test the Track instantiation."""
    track_id = 1
    track_token = "fake"  # noqa: S105
    track_duration = 120
    track_artist = "Jimi Hendrix"
    track_title = "All along the watchtower"
    track_version = "(Dylan remix)"
    track_album = "Experience"
    track_picture = "ABCDEF"
    track_filesize_mp3_128 = 128
    track_filesize_mp3_320 = 320
    track_filesize_flac = 7142

    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=DeezerSongResponseFactory.build(
            error={},
            results=DeezerSongFactory.build(
                SNG_ID=track_id,
                TRACK_TOKEN=track_token,
                DURATION=track_duration,
                ART_NAME=track_artist,
                SNG_TITLE=track_title,
                VERSION=track_version,
                ALB_TITLE=track_album,
                ALB_PICTURE=track_picture,
                FILESIZE_MP3_128=track_filesize_mp3_128,
                FILESIZE_MP3_320=track_filesize_mp3_320,
                FILESIZE_FLAC=track_filesize_flac,
            ),
        ).model_dump(),
    )

    track = Track(client=configured_onzr.deezer, track_id=track_id, background=False)

    assert track.track_id == track_id
    assert track.key == b"4den4:}:g,#j3i`a"
    assert track.status == TrackStatus.IDLE
    assert track.streamed == 0
    assert track.track_info == TrackInfo(
        id=track_id,
        token=track_token,
        duration=track_duration,
        artist=track_artist,
        title=f"{track_title} {track_version}",
        album=track_album,
        picture=track_picture,
        formats=[
            StreamQuality.MP3_128,
            StreamQuality.MP3_320,
            StreamQuality.FLAC,
        ],
    )
    assert track.token == track_token
    assert track.duration == track_duration
    assert track.artist == track_artist
    assert track.title == f"{track_title} {track_version}"
    assert track.album == track_album
    assert track.picture == track_picture
    assert track.cover_small == HttpUrl(
        "https://e-cdns-images.dzcdn.net/images/cover/ABCDEF/56x56-000000-80-0-0.jpg"
    )
    assert track.cover_medium == HttpUrl(
        "https://e-cdns-images.dzcdn.net/images/cover/ABCDEF/250x250-000000-80-0-0.jpg"
    )
    assert track.cover_big == HttpUrl(
        "https://e-cdns-images.dzcdn.net/images/cover/ABCDEF/500x500-000000-80-0-0.jpg"
    )
    assert track.cover_xl == HttpUrl(
        "https://e-cdns-images.dzcdn.net/images/cover/ABCDEF/1000x1000-000000-80-0-0.jpg"
    )
    assert (
        track.full_title
        == f"{track_artist} - {track_title} {track_version} [{track_album}]"
    )

    # If picture is None
    track.track_info.picture = None
    assert track.cover_small is None
    assert track.cover_medium is None
    assert track.cover_big is None
    assert track.cover_xl is None
    assert track.formats == [
        StreamQuality.MP3_128,
        StreamQuality.MP3_320,
        StreamQuality.FLAC,
    ]

    # Test available formats
    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=DeezerSongResponseFactory.build(
            error={},
            results=DeezerSongFactory.build(
                SNG_ID=track_id,
                TRACK_TOKEN=track_token,
                DURATION=track_duration,
                ART_NAME=track_artist,
                SNG_TITLE=track_title,
                ALB_TITLE=track_album,
                ALB_PICTURE=track_picture,
                FILESIZE_MP3_128=track_filesize_mp3_128,
                FILESIZE_MP3_320=track_filesize_mp3_320,
                FILESIZE_FLAC=0,
            ),
        ).model_dump(),
    )
    track = Track(client=configured_onzr.deezer, track_id=track_id, background=False)

    assert track.formats == [
        StreamQuality.MP3_128,
        StreamQuality.MP3_320,
    ]

    # Test when none of configured formats are available
    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=DeezerSongResponseFactory.build(
            error={},
            results=DeezerSongFactory.build(
                SNG_ID=track_id,
                TRACK_TOKEN=track_token,
                DURATION=track_duration,
                ART_NAME=track_artist,
                SNG_TITLE=track_title,
                VERSION="",
                ALB_TITLE=track_album,
                ALB_PICTURE=track_picture,
                FILESIZE_MP3_128=0,
                FILESIZE_MP3_320=0,
                FILESIZE_FLAC=0,
            ),
        ).model_dump(),
    )
    with pytest.raises(
        DeezerTrackException,
        match=r"No available formats detected for track \d+$",
    ):
        Track(client=configured_onzr.deezer, track_id=track_id, background=False)

    # Test when no version is supplied (empty string)
    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=DeezerSongResponseFactory.build(
            error={},
            results=DeezerSongFactory.build(
                SNG_ID=track_id,
                TRACK_TOKEN=track_token,
                DURATION=track_duration,
                ART_NAME=track_artist,
                SNG_TITLE=track_title,
                VERSION="",
                ALB_TITLE=track_album,
                ALB_PICTURE=track_picture,
                FILESIZE_MP3_128=track_filesize_mp3_128,
                FILESIZE_MP3_320=track_filesize_mp3_320,
                FILESIZE_FLAC=track_filesize_flac,
            ),
        ).model_dump(),
    )

    track = Track(client=configured_onzr.deezer, track_id=track_id, background=False)
    assert track.title == track_title

    # Test when no version is supplied (field not in payload)
    payload = DeezerSongResponseFactory.build(
        error={},
        results=DeezerSongFactory.build(
            SNG_ID=track_id,
            TRACK_TOKEN=track_token,
            DURATION=track_duration,
            ART_NAME=track_artist,
            SNG_TITLE=track_title,
            ALB_TITLE=track_album,
            ALB_PICTURE=track_picture,
            FILESIZE_MP3_128=track_filesize_mp3_128,
            FILESIZE_MP3_320=track_filesize_mp3_320,
            FILESIZE_FLAC=track_filesize_flac,
        ),
    )
    del payload.results.VERSION

    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=payload.model_dump(),
    )

    track = Track(client=configured_onzr.deezer, track_id=track_id, background=False)
    assert track.title == track_title


def test_track_query_quality(configured_onzr, responses):
    """Test the track query_quality method."""
    track_id = 1
    track_token = "fake"  # noqa: S105
    track_duration = 120
    track_artist = "Jimi Hendrix"
    track_title = "All along the watchtower"
    track_album = "Experience"
    track_picture = "ABCDEF"
    track_filesize_mp3_128 = 128
    track_filesize_mp3_320 = 320

    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=DeezerSongResponseFactory.build(
            error={},
            results=DeezerSongFactory.build(
                SNG_ID=track_id,
                TRACK_TOKEN=track_token,
                DURATION=track_duration,
                ART_NAME=track_artist,
                SNG_TITLE=track_title,
                ALB_TITLE=track_album,
                ALB_PICTURE=track_picture,
                FILESIZE_MP3_128=track_filesize_mp3_128,
                FILESIZE_MP3_320=track_filesize_mp3_320,
                FILESIZE_FLAC=0,
            ),
        ).model_dump(),
    )
    track = Track(client=configured_onzr.deezer, track_id=track_id, background=False)

    assert track.query_quality(StreamQuality.MP3_128) == StreamQuality.MP3_128
    assert track.query_quality(StreamQuality.MP3_320) == StreamQuality.MP3_320
    assert track.query_quality(StreamQuality.FLAC) == StreamQuality.MP3_320


def test_track_serialize(configured_onzr, responses):
    """Test the Track serialization."""
    track_id = 1
    track_token = "fake"  # noqa: S105
    track_duration = 120
    track_artist = "Jimi Hendrix"
    track_title = "All along the watchtower"
    track_album = "Experience"
    track_picture = "ABCDEF"

    responses.post(
        "http://www.deezer.com/ajax/gw-light.php",
        status=200,
        json=DeezerSongResponseFactory.build(
            error={},
            results=DeezerSongFactory.build(
                SNG_ID=track_id,
                TRACK_TOKEN=track_token,
                DURATION=track_duration,
                ART_NAME=track_artist,
                SNG_TITLE=track_title,
                VERSION="",
                ALB_TITLE=track_album,
                ALB_PICTURE=track_picture,
            ),
        ).model_dump(),
    )

    track = Track(client=configured_onzr.deezer, track_id=track_id, background=False)

    assert track.serialize() == TrackShort(
        id=track_id,
        title=track_title,
        album=track_album,
        artist=track_artist,
    )



"""Onzr: core module."""

import logging
import random
from functools import cached_property
from time import sleep
from typing import List

from vlc import Instance, MediaList, MediaListPlayer

from onzr.config import get_settings

from .deezer import DeezerClient, Track
from .models import QueuedTrack, QueuedTracks, QueueState, ServerState

logger = logging.getLogger(__name__)


class Queue:
    """Onzr playing queue."""

    def __init__(self, player: MediaListPlayer) -> None:
        """Instantiate the tracks queue."""
        self.playing: int | None = None
        self.tracks: List[Track] = []
        self.player: MediaListPlayer = player
        self.playlist: MediaList = self._activate_new_playlist(self.player)

    def __len__(self):
        """Get queue length."""
        return len(self.tracks)

    def __getitem__(self, index: int) -> Track:
        """Get track from its queue index."""
        return self.tracks[index]

    @cached_property
    def vlc_instance(self):
        """Get VLC instance."""
        return self.player.get_instance()

    @property
    def is_empty(self):
        """Check if tracks are queued."""
        return len(self) == 0

    @property
    def current(self) -> Track | None:
        """Get the current track."""
        if self.playing is None:
            return None
        return self.tracks[self.playing]

    @property
    def state(self) -> QueueState:
        """Get queue state."""
        return QueueState(playing=self.playing, queued=len(self))

    def _activate_new_playlist(self, player: MediaListPlayer) -> MediaList:
        """Create a new playlist and activate the media player with it."""
        playlist = self.vlc_instance.media_list_new()
        player.set_media_list(playlist)
        return playlist

    def add(self, tracks: List[Track]):
        """Add one or more tracks to queue."""
        start = len(self)
        self.tracks.extend(tracks)

        # Add track streaming url to the playlist
        vlc_instance = self.playlist.get_instance()
        settings = get_settings()
        for rank in range(start, start + len(tracks), 1):
            media = vlc_instance.media_new(settings.TRACK_STREAM_URL.format(rank=rank))
            self.playlist.add_media(media)

    def clear(self):
        """Empty queue."""
        self.playing = None
        self.tracks = []

        # Player-related part
        if self.playlist:
            self.playlist.release()
        self.playlist = self._activate_new_playlist(self.player)

    def shuffle(self):
        """Shuffle current track list."""
        random.shuffle(self.tracks)

    def serialize(self) -> QueuedTracks:
        """Serialize queue."""
        return QueuedTracks(
            playing=self.playing,
            tracks=[
                QueuedTrack(current=self.playing == p, position=p, track=t.serialize())
                for p, t in enumerate(self.tracks)
            ],
        )


class Onzr:
    """Onzr main class that communicates with every components.

    - deezer: Deezer API client
    - player: VLC player
    - queue: Queue instance
    """

    def __init__(self) -> None:
        """Instantiate components."""
        self.settings = get_settings()

        # Deezer API client
        self.deezer: DeezerClient = DeezerClient(
            arl=self.settings.ARL,
            blowfish=self.settings.DEEZER_BLOWFISH_SECRET,
            fast=False,
        )

        # Player
        vlc_instance: Instance = Instance()
        self.player: MediaListPlayer = vlc_instance.media_list_player_new()

        # Queue
        self.queue: Queue = Queue(player=self.player)

    def state(self) -> ServerState:
        """Get Onzr state."""
        # Wait a bit before returning the server/player state, since after performing
        # a control action on the player, its state may need some time to get updated.
        sleep(self.settings.STATE_DELAY)

        return ServerState(player=str(self.player.get_state()), queue=self.queue.state)



"""Onzr client tests."""

from time import sleep

from onzr.client import OnzrClient
from onzr.models import (
    PlayerControl,
    PlayerState,
    PlayingState,
    QueuedTracks,
    QueueState,
    ServerMessage,
    ServerState,
)

from .factories import DeezerSongFactory, DeezerSongResponseFactory


def test_queue_add(test_server, responses):
    """Test the `queue_add` method."""
    client = OnzrClient()

    track_ids = [1, 2, 3]
    for track_id in track_ids:
        responses.post(
            "http://www.deezer.com/ajax/gw-light.php",
            status=200,
            json=DeezerSongResponseFactory.build(
                error={}, results=DeezerSongFactory.build(SNG_ID=track_id)
            ).model_dump(),
        )

    assert client.queue_add([str(t) for t in track_ids]) == ServerMessage(
        message="Added 3 track(s) to queue"
    )


def test_queue_clear(test_server, configured_onzr, track):
    """Test the `queue_add` method."""
    # Empty queue
    client = OnzrClient()
    assert client.queue_clear() == ServerState(
        player="State.Stopped", queue=QueueState(playing=None, queued=0)
    )
    assert len(configured_onzr.queue) == 0
    assert configured_onzr.queue.is_empty

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    assert client.queue_clear() == ServerState(
        player="State.Stopped", queue=QueueState(playing=None, queued=0)
    )
    assert len(configured_onzr.queue) == 0
    assert configured_onzr.queue.is_empty


def test_queue_list(test_server, configured_onzr, track):
    """Test the `queue_list` method."""
    # Empty queue
    client = OnzrClient()
    assert client.queue_list() == QueuedTracks(playing=None, tracks=[])

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    queued_tracks = client.queue_list()
    assert isinstance(queued_tracks, QueuedTracks)
    assert queued_tracks.playing is None
    assert len(queued_tracks.tracks) == len(track_ids)


def test_now_playing(test_server, configured_onzr, track):
    """Test the `now_playing` method."""
    client = OnzrClient()

    # Empty queue
    assert client.now_playing() == PlayingState(
        player=PlayerState(
            state="State.NothingSpecial", length=-1, time=-1, position=-1.0
        ),
        track=None,
    )

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    configured_onzr.player.play()

    assert client.now_playing() == PlayingState(
        player=PlayerState(state="State.Opening", length=0, time=0, position=0.0),
        track=None,
    )

    # Stop the player
    configured_onzr.player.stop()


def test_state(test_server, configured_onzr, track):
    """Test the `state` method."""
    client = OnzrClient()

    # Empty queue
    assert client.state() == ServerState(
        player="State.NothingSpecial", queue=QueueState(playing=None, queued=0)
    )

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Start playing
    configured_onzr.player.play()

    assert client.state() == ServerState(
        player="State.Opening", queue=QueueState(playing=None, queued=3)
    )

    # Stop the player
    configured_onzr.player.stop()


def test_ping_server_down(configured_onzr):
    """Test the `ping` method when the server is down."""
    client = OnzrClient()
    assert client.ping() is False


def test_ping_server(test_server, configured_onzr):
    """Test the `ping` method when the server is down."""
    client = OnzrClient()
    assert client.ping()


def test_play(test_server, configured_onzr, track):
    """Test the `play` method."""
    client = OnzrClient()

    # Empty queue
    assert client.play() == PlayerControl(
        action="play",
        state=ServerState(
            player="State.NothingSpecial", queue=QueueState(playing=None, queued=0)
        ),
    )

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Start playing: should be opening the file
    assert client.play() == PlayerControl(
        action="play",
        state=ServerState(
            player="State.Opening", queue=QueueState(playing=None, queued=3)
        ),
    )

    # Wait a bit and we should really be playing something
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=0, queued=3)
    )

    # Start playing from a queue index (aka rank)
    assert client.play(rank=1) == PlayerControl(
        action="play",
        state=ServerState(
            player="State.Opening", queue=QueueState(playing=0, queued=3)
        ),
    )

    # Wait a bit and we should have switched to rank 1 track
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=1, queued=3)
    )


def test_pause(test_server, configured_onzr, track):
    """Test the `pause` method."""
    client = OnzrClient()

    # Empty queue
    assert client.pause() == PlayerControl(
        action="pause",
        state=ServerState(
            player="State.NothingSpecial", queue=QueueState(playing=None, queued=0)
        ),
    )

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Start playing: should be opening the file
    configured_onzr.player.play()

    # Wait a bit and we should really be playing something
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=0, queued=3)
    )

    # Start playing from a queue index (aka rank)
    assert client.pause() == PlayerControl(
        action="pause",
        state=ServerState(player="State.Paused", queue=QueueState(playing=0, queued=3)),
    )

    # Toggle pause
    assert client.pause() == PlayerControl(
        action="pause",
        state=ServerState(
            player="State.Playing", queue=QueueState(playing=0, queued=3)
        ),
    )


def test_stop(test_server, configured_onzr, track):
    """Test the `stop` method."""
    client = OnzrClient()

    # Empty queue
    assert client.stop() == PlayerControl(
        action="stop",
        state=ServerState(
            player="State.Stopped", queue=QueueState(playing=None, queued=0)
        ),
    )

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Start playing: should be opening the file
    configured_onzr.player.play()

    # Wait a bit and we should really be playing something
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=0, queued=3)
    )

    # Stop playing
    assert client.stop() == PlayerControl(
        action="stop",
        state=ServerState(
            player="State.Stopped", queue=QueueState(playing=0, queued=3)
        ),
    )


def test_next(test_server, configured_onzr, track):
    """Test the `next` method."""
    client = OnzrClient()

    # Empty queue
    assert client.next() == PlayerControl(
        action="next",
        state=ServerState(
            player="State.NothingSpecial", queue=QueueState(playing=None, queued=0)
        ),
    )

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Should start playing first track in queue
    assert client.next() == PlayerControl(
        action="next",
        state=ServerState(
            player="State.Opening", queue=QueueState(playing=None, queued=3)
        ),
    )

    # Wait a bit and we should really be playing something
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=0, queued=3)
    )

    # Next track
    assert client.next() == PlayerControl(
        action="next",
        state=ServerState(
            player="State.Opening", queue=QueueState(playing=0, queued=3)
        ),
    )

    # Wait a bit and we should be playing the next track
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=1, queued=3)
    )


def test_previous(test_server, configured_onzr, track):
    """Test the `previous` method."""
    client = OnzrClient()

    # Empty queue
    assert client.previous() == PlayerControl(
        action="previous",
        state=ServerState(
            player="State.NothingSpecial", queue=QueueState(playing=None, queued=0)
        ),
    )

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Should start playing first track in queue
    assert client.previous() == PlayerControl(
        action="previous",
        state=ServerState(
            player="State.Opening", queue=QueueState(playing=None, queued=3)
        ),
    )
    # Wait a bit and we should really be playing something
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=0, queued=3)
    )

    # Previous should restart current track from the beginning
    assert client.previous() == PlayerControl(
        action="previous",
        state=ServerState(
            player="State.Playing", queue=QueueState(playing=0, queued=3)
        ),
    )
    # Wait a bit and we should be playing the previous track
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=0, queued=3)
    )

    # Play track 2
    configured_onzr.player.play_item_at_index(2)
    assert client.previous() == PlayerControl(
        action="previous",
        state=ServerState(
            player="State.Opening", queue=QueueState(playing=0, queued=3)
        ),
    )
    # Wait a bit and we should be playing the previous track
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=1, queued=3)
    )

    # Again?
    assert client.previous() == PlayerControl(
        action="previous",
        state=ServerState(
            player="State.Opening", queue=QueueState(playing=1, queued=3)
        ),
    )
    # Wait a bit and we should be playing the previous track
    sleep(0.1)
    assert client.state() == ServerState(
        player="State.Playing", queue=QueueState(playing=0, queued=3)
    )



"""Onzr: http server."""

import logging
from functools import lru_cache
from typing import Annotated, List

from fastapi import Depends, FastAPI, HTTPException, Path, status
from fastapi.responses import StreamingResponse

from .config import get_settings
from .core import Onzr
from .deezer import Track
from .models import (
    PlayerControl,
    PlayerState,
    PlayingState,
    PlayQueryParams,
    QueuedTracks,
    ServerMessage,
    ServerState,
)

logger = logging.getLogger(__name__)

settings = get_settings()

app = FastAPI(title="Onzr", root_path=settings.API_ROOT_URL, debug=settings.DEBUG)


@lru_cache
def get_onzr() -> Onzr:
    """Get Onzr core instance."""
    return Onzr()


# --- Routes


@app.post("/queue/")
async def queue_add(
    onzr: Annotated[Onzr, Depends(get_onzr)],
    track_ids: List[int],
) -> ServerMessage:
    """Add tracks to queue given their identifiers."""
    tracks = [Track(onzr.deezer, id_, background=True) for id_ in track_ids]
    onzr.queue.add(tracks=tracks)
    return ServerMessage(message=f"Added {len(tracks)} track(s) to queue")


@app.delete("/queue/")
async def queue_clear(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> ServerState:
    """Clear tracks queue."""
    onzr.player.stop()
    onzr.queue.clear()
    return onzr.state()


@app.get("/queue/")
async def queue_list(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> QueuedTracks:
    """List queue tracks."""
    return onzr.queue.serialize()


@app.get(settings.TRACK_STREAM_ENDPOINT)
async def stream_track(
    onzr: Annotated[Onzr, Depends(get_onzr)],
    rank: Annotated[int, Path(title="Track queue rank")],
) -> StreamingResponse:
    """Stream Deezer track given its identifer."""
    if onzr.queue.is_empty:
        raise HTTPException(
            status_code=status.HTTP_204_NO_CONTENT, detail="Queue is empty."
        )
    if rank >= len(onzr.queue):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Track rank out of range."
        )
    onzr.queue.playing = rank
    track = onzr.queue[rank]
    # Refresh track token in case it expired
    track.refresh()
    quality = track.query_quality(settings.QUALITY)
    return StreamingResponse(track.stream(quality), media_type=quality.media_type)


@app.get("/now")
async def now_playing(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> PlayingState:
    """Get info about current track."""
    track = onzr.queue.current
    media_player = onzr.player.get_media_player()
    return PlayingState(
        player=PlayerState(
            state=str(media_player.get_state()),
            length=media_player.get_length(),
            time=media_player.get_time(),
            position=media_player.get_position(),
        ),
        track=track.serialize() if track else None,
    )


@app.post("/play")
async def play(
    onzr: Annotated[Onzr, Depends(get_onzr)], params: PlayQueryParams
) -> PlayerControl:
    """Start playing current queue."""
    if params.rank is not None:
        # TODO: rank should be < len(queue)
        onzr.player.play_item_at_index(params.rank)
    else:
        onzr.player.play()
    return PlayerControl(action="play", state=onzr.state())


@app.post("/pause")
async def pause(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> PlayerControl:
    """Pause/resume playing."""
    onzr.player.pause()
    return PlayerControl(action="pause", state=onzr.state())


@app.post("/stop")
async def stop(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> PlayerControl:
    """Stop playing."""
    onzr.player.stop()
    return PlayerControl(action="stop", state=onzr.state())


@app.post("/next")
async def next(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> PlayerControl:
    """Play next track in queue."""
    onzr.player.next()
    return PlayerControl(action="next", state=onzr.state())


@app.post("/previous")
async def previous(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> PlayerControl:
    """Play previous track in queue."""
    onzr.player.previous()
    return PlayerControl(action="previous", state=onzr.state())


@app.get("/state")
async def state(
    onzr: Annotated[Onzr, Depends(get_onzr)],
) -> ServerState:
    """Server state."""
    return onzr.state()


@app.get("/ping")
async def ping() -> None:
    """Server ping."""
    return



"""Onzr CLI tests."""

import copy
import datetime
import json
import logging
import re
from os import stat
from pathlib import Path
from time import sleep
from unittest.mock import MagicMock, patch

import click
import pytest
import uvicorn
import vlc
import yaml

import onzr
from onzr.cli import ExitCodes, cli
from onzr.deezer import DeezerClient
from onzr.exceptions import OnzrConfigurationError
from onzr.models import AlbumShort, ArtistShort, Collection, TrackShort
from tests.factories import DeezerSongFactory, DeezerSongResponseFactory

# Test fixtures
artist_1 = ArtistShort(id=1, name="foo")
artist_2 = ArtistShort(id=2, name="bar")
artists_collection: Collection = [artist_1, artist_2]
album_1 = AlbumShort(
    id=11,
    title="foo",
    artist="foo",
    release_date=datetime.date(2025, 1, 1).isoformat(),
)
album_2 = AlbumShort(
    id=12,
    title="bar",
    artist="bar",
    release_date=datetime.date(1925, 10, 1).isoformat(),
)
albums_collection: Collection = [album_1, album_2]
track_1 = TrackShort(id=21, title="foo", album="foo", artist="foo")
track_2 = TrackShort(id=22, title="bar", album="bar", artist="foo")
tracks_collection: Collection = [track_1, track_2]

# System exit codes
SYSTEM_EXIT_1 = 1
SYSTEM_EXIT_2 = 2


def test_require_server_command_wrapper(configured_cli_runner):
    """Test the `require_server` decorator."""
    result = configured_cli_runner.invoke(cli, ["state"])
    assert result.exit_code == ExitCodes.SERVER_DOWN
    assert "❌ Onzr server is down, run `onzr serve` first." in result.stdout


def test_command_help(cli_runner):
    """Test the `onzr --help` command."""
    result = cli_runner.invoke(cli, ["--help"])
    assert result.exit_code == ExitCodes.OK


def test_openapi_command(test_server, configured_cli_runner):
    """Test the `onzr openapi` command."""
    result = configured_cli_runner.invoke(cli, ["openapi"])
    assert result.exit_code == ExitCodes.OK

    # FIXME: a server log is captured as stdout (row 1)
    schema = json.loads(result.stdout.split("\n")[1])
    assert "openapi" in schema


def test_init_command_without_input(cli_runner, settings_file):
    """Test the `onzr init` command without ARL input."""
    assert settings_file.exists() is False

    # No ARL setting is provided
    result = cli_runner.invoke(cli, ["init"])
    assert result.exit_code == SYSTEM_EXIT_1

    # Base configuration exists but without ARL setting
    assert settings_file.exists() is True

    # Dist file and its initial copy hould be identical
    dist = Path(onzr.__file__).parent / Path(f"{settings_file.name}.dist")
    assert dist.exists()
    assert dist.read_text() == settings_file.read_text()


def test_init_command(cli_runner, settings_file):
    """Test the `onzr init` command."""
    assert settings_file.exists() is False

    result = cli_runner.invoke(cli, ["init"], input="fake-arl")
    assert result.exit_code == ExitCodes.OK

    # Settings file should exist now
    assert settings_file.exists() is True

    # SETTINGS_FILE should be updated compared to the distributed template
    settings_dist = Path(onzr.__file__).parent / Path(f"{settings_file.name}.dist")
    assert settings_dist.exists()
    settings_file_content = settings_file.read_text()
    assert settings_dist.read_text() != settings_file_content
    assert re.search("^ARL: .*", settings_file_content)


def test_init_command_does_not_overwrite_settings(cli_runner, settings_file):
    """Test the `onzr init` command does not overwrite existing settings."""
    assert settings_file.exists() is False

    # Get most recent modification time
    result = cli_runner.invoke(cli, ["init"], input="fake-arl")
    original_stat = stat(settings_file)

    # Re-run the `init` command without reset mode shouldn't touch settings
    result = cli_runner.invoke(cli, ["init"], input="fake-arl")
    assert result.exception is not None
    assert isinstance(result.exception, OnzrConfigurationError)
    assert (
        result.exception.args[0]
        == f"Configuration file '{settings_file}' already exists!"
    )
    assert original_stat == stat(settings_file)


def test_config_command_no_config_file(cli_runner, settings_file):
    """Test the `onzr config` command when the configuration file does not exist."""
    assert settings_file.exists() is False

    result = cli_runner.invoke(cli, ["config"])
    assert result.exit_code == ExitCodes.INCOMPLETE_CONFIGURATION
    assert "Configuration file does not exist, use `onzr init` first." in result.stdout


def test_config_command(configured_cli_runner, settings_file):
    """Test the `onzr config` command."""
    assert settings_file.exists() is True

    # Mock click.edit function
    click.edit = MagicMock(return_value=0)

    # Get configuration file path
    result = configured_cli_runner.invoke(cli, ["config", "-p"])
    assert result.exit_code == ExitCodes.OK
    assert result.stdout.split("\n")[0] == str(settings_file)
    click.edit.assert_not_called()

    # Display configuration
    result = configured_cli_runner.invoke(cli, ["config"])
    assert result.exit_code == ExitCodes.OK
    with settings_file.open() as s:
        assert yaml.safe_load(result.stdout) == yaml.safe_load(s)
    click.edit.assert_not_called()

    # Edit configuration
    result = configured_cli_runner.invoke(cli, ["config", "-e"])
    assert result.exit_code == ExitCodes.OK
    click.edit.assert_called_once()


def test_search_command_with_no_argument(configured_cli_runner):
    """Test the `onzr search` without any argument."""
    result = configured_cli_runner.invoke(cli, ["search"])
    assert result.exit_code == ExitCodes.NOT_FOUND


@pytest.mark.parametrize("option", ("artist", "album", "track"))
def test_search_command_with_no_match(configured_cli_runner, monkeypatch, option):
    """Test the `onzr search` command with no match."""

    def search(*args, **kwargs):
        """Monkeypatch search."""
        return []

    monkeypatch.setattr(DeezerClient, "search", search)

    result = configured_cli_runner.invoke(cli, ["search", f"--{option}", "foo"])
    assert result.exit_code == ExitCodes.NOT_FOUND


@pytest.mark.parametrize(
    "option,results",
    (
        ("artist", artists_collection),
        ("album", albums_collection),
        ("track", tracks_collection),
    ),
)
def test_search_command(configured_cli_runner, monkeypatch, option, results):
    """Test the `onzr search` command."""

    def search(*args, **kwargs):
        """Monkeypatch search."""
        return results

    monkeypatch.setattr(DeezerClient, "search", search)

    result = configured_cli_runner.invoke(cli, ["search", f"--{option}", "foo"])
    assert result.exit_code == ExitCodes.OK

    # Test ids option
    result = configured_cli_runner.invoke(
        cli, ["search", f"--{option}", "foo", "--ids"]
    )
    assert result.exit_code == ExitCodes.OK
    expected = "".join([f"{r.id}\n" for r in results])
    assert result.stdout == expected

    # Test first option
    result = configured_cli_runner.invoke(
        cli, ["search", f"--{option}", "foo", "--ids", "--first"]
    )
    assert result.exit_code == ExitCodes.OK
    expected = f"{results[0].id}\n"
    assert result.stdout == expected


def test_artist_command_with_no_id(configured_cli_runner):
    """Test the `onzr artist` command with no ID."""
    result = configured_cli_runner.invoke(cli, ["artists"])
    assert result.exit_code == SYSTEM_EXIT_2


def test_artist_command(configured_cli_runner, monkeypatch):
    """Test the `onzr artist` command."""
    # One should choose one type of result
    result = configured_cli_runner.invoke(cli, ["artist", "1", "--no-top"])
    assert result.exit_code == ExitCodes.INVALID_ARGUMENTS

    top_collection = copy.copy(tracks_collection)
    top_collection.reverse()

    def artist(*args, **kwargs):
        """Monkeypatch artist."""
        if kwargs.get("radio"):
            return tracks_collection
        elif kwargs.get("top"):
            return top_collection
        elif kwargs.get("albums"):
            return albums_collection

    monkeypatch.setattr(DeezerClient, "artist", artist)

    # Default using an argument
    result = configured_cli_runner.invoke(cli, ["artist", "1"])
    assert result.exit_code == ExitCodes.OK
    result = configured_cli_runner.invoke(cli, ["artist", "--ids", "1"])
    assert result.exit_code == ExitCodes.OK
    assert result.stdout == "".join([f"{t.id}\n" for t in top_collection])

    # Top using an argument
    result = configured_cli_runner.invoke(cli, ["artist", "--ids", "--top", "1"])
    assert result.exit_code == ExitCodes.OK
    assert result.stdout == "".join([f"{t.id}\n" for t in top_collection])

    # Top using stdin
    for input in ["1", " 1", " 1 ", "1 "]:
        result = configured_cli_runner.invoke(
            cli, ["artist", "--ids", "--top", "-"], input=input
        )
        assert result.exit_code == ExitCodes.OK
        assert result.stdout == "".join([f"{t.id}\n" for t in top_collection])

    # Radio
    result = configured_cli_runner.invoke(cli, ["artist", "--ids", "--radio", "1"])
    assert result.exit_code == ExitCodes.OK
    assert result.stdout == "".join([f"{t.id}\n" for t in tracks_collection])

    # Albums
    result = configured_cli_runner.invoke(cli, ["artist", "--ids", "--albums", "1"])
    assert result.exit_code == ExitCodes.OK
    assert result.stdout == "".join([f"{a.id}\n" for a in albums_collection])


def test_album_command(configured_cli_runner, monkeypatch):
    """Test the `onzr album` command."""

    def album(*args, **kwargs):
        """Monkeypatch album."""
        return tracks_collection

    monkeypatch.setattr(DeezerClient, "album", album)

    # Standard run
    result = configured_cli_runner.invoke(cli, ["album", "1"])
    assert result.exit_code == ExitCodes.OK

    # Display only track ids
    result = configured_cli_runner.invoke(cli, ["album", "--ids", "1"])
    assert result.exit_code == ExitCodes.OK
    assert result.stdout == "".join([f"{t.id}\n" for t in tracks_collection])

    # Use stdin
    for input in ["1", " 1", " 1 ", "1 "]:
        result = configured_cli_runner.invoke(cli, ["album", "--ids", "-"], input=input)
        assert result.exit_code == ExitCodes.OK
        assert result.stdout == "".join([f"{t.id}\n" for t in tracks_collection])


def test_mix_command(configured_cli_runner, monkeypatch):
    """Test the `onzr mix` command."""

    def search(*args, **kwargs):
        """Monkeypatch search."""
        return artists_collection

    monkeypatch.setattr(DeezerClient, "search", search)

    track_3 = TrackShort(id="31", title="lol", album="foo", artist="foo")
    track_4 = TrackShort(id="32", title="doe", album="bar", artist="bar")
    deep_collection: Collection = [track_3, track_4]

    def artist(*args, **kwargs):
        """Monkeypatch artist."""
        if kwargs.get("radio"):
            return deep_collection
        elif kwargs.get("top"):
            return tracks_collection

    monkeypatch.setattr(DeezerClient, "artist", artist)

    # Standard mix
    result = configured_cli_runner.invoke(cli, ["mix", "foo", "bar"])
    assert result.exit_code == ExitCodes.OK

    result = configured_cli_runner.invoke(cli, ["mix", "foo", "bar", "--ids"])
    assert result.exit_code == ExitCodes.OK
    # As tracks are shuffled, we need to sort them
    assert sorted(result.stdout.split()) == sorted(
        [f"{t.id}" for t in tracks_collection] * 2
    )

    # Deep mix
    result = configured_cli_runner.invoke(cli, ["mix", "foo", "bar", "--deep"])
    assert result.exit_code == ExitCodes.OK

    result = configured_cli_runner.invoke(cli, ["mix", "foo", "bar", "--ids", "--deep"])
    assert result.exit_code == ExitCodes.OK
    # As tracks are shuffled, we need to sort them
    assert sorted(result.stdout.split()) == sorted(
        [f"{t.id}" for t in deep_collection] * 2
    )


def test_add_command(test_server, responses, configured_cli_runner):
    """Test the `onzr add` command."""
    track_ids = [1, 2, 3]
    for track_id in track_ids:
        responses.post(
            "http://www.deezer.com/ajax/gw-light.php",
            status=200,
            json=DeezerSongResponseFactory.build(
                error={}, results=DeezerSongFactory.build(SNG_ID=track_id)
            ).model_dump(),
        )

    result = configured_cli_runner.invoke(cli, ["add", "1", "2", "3"])
    assert result.exit_code == ExitCodes.OK
    assert "Added 3 track(s) to queue" in result.stdout

    # Pass ids via stdin
    result = configured_cli_runner.invoke(
        cli, ["add", "-"], input="\n".join(map(str, track_ids))
    )
    assert result.exit_code == ExitCodes.OK
    assert "Added 3 track(s) to queue" in result.stdout


def test_queue_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr queue` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["queue"])
    assert result.exit_code == ExitCodes.OK
    assert "Queue is empty, use onzr add to start adding tracks." in result.stdout

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    result = configured_cli_runner.invoke(cli, ["queue"])
    assert result.exit_code == ExitCodes.OK
    assert all(x in result.stdout for x in [f"🧵   {i}" for i in range(1, 4)])


def test_clear_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr clear` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["clear"])
    assert result.exit_code == ExitCodes.OK
    assert "📢 Player: Stopped · Queue: None / 0" in result.stdout

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    result = configured_cli_runner.invoke(cli, ["clear"])
    assert result.exit_code == ExitCodes.OK
    assert "📢 Player: Stopped · Queue: None / 0" in result.stdout
    assert configured_onzr.queue.is_empty


def test_now_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr clear` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["now"])
    assert result.exit_code == ExitCodes.OK
    assert "Nothing more has been queued" in result.stdout

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Still nothing to display
    result = configured_cli_runner.invoke(cli, ["now"])
    assert result.exit_code == ExitCodes.OK
    assert "Nothing more has been queued" in result.stdout

    # Start playing
    configured_onzr.player.play()
    sleep(0.3)
    result = configured_cli_runner.invoke(cli, ["now"])
    assert result.exit_code == ExitCodes.OK
    assert "· (1/3)" in result.stdout
    assert "▶️" in result.stdout
    assert "Next:" in result.stdout


def test_play_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr play` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["play"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.NothingSpecial

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Play
    result = configured_cli_runner.invoke(cli, ["play"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.player.get_state() == vlc.State.Opening

    # Play an invalid track rank
    result = configured_cli_runner.invoke(cli, ["play", "--rank", "0"])
    assert result.exit_code == ExitCodes.INVALID_ARGUMENTS
    assert "Invalid rank" in result.stdout

    # Play the first track in queue
    result = configured_cli_runner.invoke(cli, ["play", "--rank", "1"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.player.get_state() == vlc.State.Opening

    # Play the second track in queue
    result = configured_cli_runner.invoke(cli, ["play", "--rank", "2"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.player.get_state() == vlc.State.Opening

    # Stop the player
    configured_onzr.player.stop()


def test_pause_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr pause` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["pause"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.NothingSpecial

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Pause without prior play event
    result = configured_cli_runner.invoke(cli, ["pause"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.NothingSpecial

    # Play
    configured_onzr.player.play()
    sleep(0.3)
    result = configured_cli_runner.invoke(cli, ["pause"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.Paused

    # Toggle
    result = configured_cli_runner.invoke(cli, ["pause"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.player.get_state() == vlc.State.Playing


def test_stop_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr stop` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["stop"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.Stopped

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Stop while playing
    configured_onzr.player.play()
    result = configured_cli_runner.invoke(cli, ["stop"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.Stopped


def test_next_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr next` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["next"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.NothingSpecial

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Play first in queue
    configured_onzr.player.play()
    sleep(0.3)
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.queue.playing == 0

    # Next
    for expected in (1, 2):
        result = configured_cli_runner.invoke(cli, ["next"])
        sleep(0.3)
        assert result.exit_code == ExitCodes.OK
        assert configured_onzr.player.is_playing() == 1
        assert configured_onzr.queue.playing == expected

    expected = 2
    result = configured_cli_runner.invoke(cli, ["next"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.queue.playing == expected


def test_previous_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr previous` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["previous"])
    assert result.exit_code == ExitCodes.OK
    assert configured_onzr.player.is_playing() == 0
    assert configured_onzr.player.get_state() == vlc.State.NothingSpecial

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    # Play first in queue
    current = 2
    configured_onzr.player.play_item_at_index(current)
    sleep(0.3)
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.queue.playing == current

    # Previous
    result = configured_cli_runner.invoke(cli, ["previous"])
    assert result.exit_code == ExitCodes.OK
    sleep(0.3)
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.queue.playing == 1

    result = configured_cli_runner.invoke(cli, ["previous"])
    assert result.exit_code == ExitCodes.OK
    sleep(0.3)
    assert configured_onzr.player.is_playing() == 1
    assert configured_onzr.queue.playing == 0


def test_serve_command(configured_cli_runner, configured_onzr):
    """Test the `onzr serve` command."""
    with (
        patch.object(uvicorn.Server, "run") as mock_run,
        patch("uvicorn.Config", spec=True) as MockConfig,
    ):
        result = configured_cli_runner.invoke(cli, ["serve"])
        assert result.exit_code == ExitCodes.OK
        MockConfig.assert_called_once_with(
            "onzr.server:app", host="localhost", port=9473, log_level=logging.INFO
        )
        assert mock_run.called

    # Test log level checking
    result = configured_cli_runner.invoke(
        cli,
        ["serve", "--log-level", "foo"],
    )
    assert result.exit_code == ExitCodes.INVALID_ARGUMENTS
    assert "Forbidden log-level" in result.stdout

    # Change defaults
    with (
        patch.object(uvicorn.Server, "run") as mock_run,
        patch("uvicorn.Config", spec=True) as MockConfig,
    ):
        result = configured_cli_runner.invoke(
            cli,
            ["serve", "--host", "127.0.0.1", "--port", 9999, "--log-level", "debug"],
        )
        assert result.exit_code == ExitCodes.OK
        MockConfig.assert_called_once_with(
            "onzr.server:app", host="127.0.0.1", port=9999, log_level=logging.DEBUG
        )
        assert mock_run.called


def test_state_command(test_server, configured_cli_runner, configured_onzr, track):
    """Test the `onzr state` command."""
    # Empty queue
    result = configured_cli_runner.invoke(cli, ["state"])
    assert result.exit_code == ExitCodes.OK
    assert "📢 Player: NothingSpecial · Queue: None / 0" in result.stdout

    # Fill the queue
    track_ids = [1, 2, 3]
    configured_onzr.queue.add([track(track_id) for track_id in track_ids])
    assert len(configured_onzr.queue) == len(track_ids)

    result = configured_cli_runner.invoke(cli, ["state"])
    assert result.exit_code == ExitCodes.OK
    assert "📢 Player: NothingSpecial · Queue: None / 3" in result.stdout

    # Start playing
    configured_onzr.player.play()
    sleep(0.3)
    result = configured_cli_runner.invoke(cli, ["state"])
    assert result.exit_code == ExitCodes.OK
    assert "📢 Player: Playing · Queue: 1 / 3" in result.stdout

    # Stop the player
    configured_onzr.player.stop()
    result = configured_cli_runner.invoke(cli, ["state"])
    assert result.exit_code == ExitCodes.OK
    assert "📢 Player: Stopped · Queue: 1 / 3" in result.stdout


def test_version_command(configured_cli_runner):
    """Test the `onzr version` command."""
    semver = (
        r"(?P<major>0|[1-9]\d*)"
        r"\.(?P<minor>0|[1-9]\d*)"
        r"\.(?P<patch>0|[1-9]\d*)"
        r"(?P<extras>.*)"
    )
    pattern = re.compile("🔖 Version: " + semver)
    result = configured_cli_runner.invoke(cli, ["version"])
    assert result.exit_code == ExitCodes.OK
    assert pattern.match(result.stdout)



"""Onzr: http client."""

from typing import Annotated, List, Optional

import requests
from annotated_types import Ge

from .config import get_settings
from .models import (
    PlayerControl,
    PlayingState,
    PlayQueryParams,
    QueuedTracks,
    ServerMessage,
    ServerState,
)


class OnzrClient:
    """Onzr API client."""

    def __init__(self):
        """Initialize Onzr API client."""
        self.http_headers = {
            "User-Agent": "Onzr Client",
        }
        self.session = requests.session()
        self.session.headers = self.http_headers

        settings = get_settings()
        self.base_url = settings.SERVER_BASE_URL
        self.ping_timeout = settings.PING_TIMEOUT

    # Queue
    def queue_add(self, track_ids: List[str]) -> ServerMessage:
        """Add tracks to queue given their identifiers."""
        response = self.session.post(f"{self.base_url}/queue/", json=track_ids)
        return ServerMessage.model_validate_json(response.text)

    def queue_clear(self) -> ServerState:
        """Clear tracks queue."""
        response = self.session.delete(f"{self.base_url}/queue/")
        return ServerState.model_validate_json(response.text)

    def queue_list(self) -> QueuedTracks:
        """List queue tracks."""
        response = self.session.get(f"{self.base_url}/queue/")
        return QueuedTracks.model_validate_json(response.text)

    # Status
    def now_playing(self) -> PlayingState:
        """Get info about current track."""
        response = self.session.get(f"{self.base_url}/now")
        return PlayingState.model_validate_json(response.text)

    def state(self) -> ServerState:
        """Get server status."""
        response = self.session.get(f"{self.base_url}/state")
        return ServerState.model_validate_json(response.text)

    def ping(self) -> bool:
        """Get server status."""
        try:
            response = self.session.get(
                f"{self.base_url}/ping", timeout=self.ping_timeout
            )
        except requests.exceptions.ConnectionError:
            return False
        return response.status_code == requests.codes.ok

    # Controls
    def play(self, rank: Optional[Annotated[int, Ge(0)]] = None) -> PlayerControl:
        """Start playing current queue."""
        params = PlayQueryParams(rank=rank)
        response = self.session.post(
            f"{self.base_url}/play",
            json=params.model_dump() if rank is not None else {},
        )
        return PlayerControl.model_validate_json(response.text)

    def pause(self) -> PlayerControl:
        """Pause/resume playing."""
        response = self.session.post(f"{self.base_url}/pause")
        return PlayerControl.model_validate_json(response.text)

    def stop(self) -> PlayerControl:
        """Stop playing."""
        response = self.session.post(f"{self.base_url}/stop")
        return PlayerControl.model_validate_json(response.text)

    def next(self) -> PlayerControl:
        """Play next track in queue."""
        response = self.session.post(f"{self.base_url}/next")
        return PlayerControl.model_validate_json(response.text)

    def previous(self) -> PlayerControl:
        """Play previous track in queue."""
        response = self.session.post(f"{self.base_url}/previous")
        return PlayerControl.model_validate_json(response.text)



"""Onzr exceptions."""


class OnzrConfigurationError(Exception):
    """Raised when onzr cannot load or create its configuration."""


class DeezerTrackException(Exception):
    """Raised when onzr cannot handle a Deezer track."""



"""Test models."""

from pydantic import BaseModel


class BaseDeezerGWResponse(BaseModel):
    """Deezer API Gateway response base Model."""

    error: dict = {}
    results: BaseModel


class DeezerSong(BaseModel):
    """Deezer API Song."""

    SNG_ID: int
    TRACK_TOKEN: str
    DURATION: int
    ART_NAME: str
    SNG_TITLE: str
    VERSION: str | None = None
    ALB_TITLE: str
    ALB_PICTURE: str
    FILESIZE_MP3_128: int
    FILESIZE_MP3_320: int
    FILESIZE_FLAC: int


class DeezerSongResponse(BaseDeezerGWResponse):
    """Deezer API Gateway Song info response."""

    results: DeezerSong
